@@ -1,4 +1,4 @@
-import os, subprocess, copy, json, tqdm, time, threading, queue, requests, abc
+import os, subprocess, copy, json, tqdm, time, threading, queue, requests, abc, logging, backoff
 import multiprocessing as mp
 import pandas as pd
 import numpy as np
@@ -10,7 +10,7 @@ from astropy import constants
 from astropy.cosmology import Planck18
 import matplotlib.pyplot as plt
 
-from timewise.general import main_logger, cache_dir, plots_dir, output_dir
+from timewise.general import main_logger, cache_dir, plots_dir, output_dir, logger_format, backoff_hndlr
 
 
 logger = main_logger.getChild(__name__)
@@ -126,6 +126,9 @@ class WISEDataBase(abc.ABC):
         "moon_masked like '00%'"
     ]
 
+    parent_wise_source_id_key = 'AllWISE_id'
+    parent_sample_wise_skysep_key = 'sep_to_WISE_source'
+
     def __init__(self,
                  base_name,
                  parent_sample_class,
@@ -144,8 +147,8 @@ class WISEDataBase(abc.ABC):
         # --------------------------- vvvv set up parent sample vvvv --------------------------- #
         self.parent_ra_key = parent_sample.default_keymap['ra'] if parent_sample else None
         self.parent_dec_key = parent_sample.default_keymap['dec'] if parent_sample else None
-        self.parent_wise_source_id_key = 'AllWISE_id'
-        self.parent_sample_wise_skysep_key = 'sep_to_WISE_source'
+        self.parent_wise_source_id_key = WISEDataBase.parent_wise_source_id_key
+        self.parent_sample_wise_skysep_key = WISEDataBase.parent_sample_wise_skysep_key
         self.parent_sample_default_entries = {
             self.parent_wise_source_id_key: "",
             self.parent_sample_wise_skysep_key: np.inf
@@ -179,6 +182,11 @@ class WISEDataBase(abc.ABC):
                 logger.debug(f"making directory {d}")
                 os.makedirs(d)
 
+        file_handler = logging.FileHandler(filename=self.cache_dir + '/log.err', mode="a")
+        file_handler.setLevel("WARNING")
+        file_handler.setFormatter(logger_format)
+        logger.addHandler(file_handler)
+
         self.submit_file = os.path.join(self.cluster_dir, 'submit.txt')
 
         # set up result attributes
@@ -222,8 +230,6 @@ class WISEDataBase(abc.ABC):
             logger.warning(f"Very large number of chunks ({value})! "
                            f"Pay attention when getting photometry to not kill IRSA!")
 
-        self._n_chunks = value
-
         if self.parent_sample:
 
             self.chunk_map = np.zeros(len(self.parent_sample.df))
@@ -233,6 +239,11 @@ class WISEDataBase(abc.ABC):
                 end_ind = start_ind + N_in_chunk
                 self.chunk_map[start_ind:end_ind] = int(i)
 
+            self._n_chunks = int(max(self.chunk_map)) + 1
+
+            if self._n_chunks != value:
+                logger.info(f"All objectes included in {self._n_chunks:.0f} chunks.")
+
         else:
             logger.warning("No parent sample given! Can not calculate dec interval masks!")
 
@@ -511,6 +522,11 @@ class WISEDataBase(abc.ABC):
 
         if chunks is None:
             chunks = list(range(round(int(self.n_chunks * perc))))
+        else:
+            cm = [c not in self.chunk_map for c in chunks]
+            if np.any(cm):
+                raise ValueError(f"Chunks {np.array(chunks)[cm]} are not in chunk map. "
+                                 f"Probably they are larger than the set chunk number of {self._n_chunks}")
 
         if service is None:
             elements_per_chunk = len(self.parent_sample.df) / self.n_chunks
@@ -546,15 +562,14 @@ class WISEDataBase(abc.ABC):
             else:
                 return os.path.join(self._cache_photometry_dir, fn + f"_{jobID}.json")
 
-    def _load_lightcurves(self, service, chunk_number=None, jobID=None, remove=False):
+    def _load_lightcurves(self, service, chunk_number=None, jobID=None, return_filename=False):
         fn = self._lightcurve_filename(service, chunk_number, jobID)
         logger.debug(f"loading {fn}")
         try:
             with open(fn, "r") as f:
                 lcs = json.load(f)
-            if remove:
-                logger.debug(f"removing {fn}")
-                os.remove(fn)
+            if return_filename:
+                return lcs, fn
             return lcs
         except FileNotFoundError:
             logger.warning(f"No file {fn}")
@@ -610,12 +625,15 @@ class WISEDataBase(abc.ABC):
             raise NotImplementedError
 
         lcs = None
+        fns = list()
         for i in itr[1]:
             kw = dict(kwargs)
             kw[itr[0]] = i
-            kw['remove'] = remove
-            ilcs = self._load_lightcurves(**kw)
-            if not isinstance(ilcs, type(None)):
+            kw['return_filename'] = True
+            res = self._load_lightcurves(**kw)
+            if not isinstance(res, type(None)):
+                ilcs, ifn = res
+                fns.append(ifn)
                 if isinstance(lcs, type(None)):
                     lcs = dict(ilcs)
                 else:
@@ -623,6 +641,10 @@ class WISEDataBase(abc.ABC):
 
         self._save_lightcurves(lcs, service=service, chunk_number=chunk_number, overwrite=overwrite)
 
+        if remove:
+            for fn in tqdm.tqdm(fns, desc="removing files"):
+                os.remove(fn)
+
     # ----------------------------------------------------------------------------------- #
     # START using GATOR to get photometry        #
     # ------------------------------------------ #
@@ -817,14 +839,23 @@ class WISEDataBase(abc.ABC):
         while True:
             if N_tries == 0:
                 logger.warning("No more tries left!")
+                raise vo.dal.exceptions.DALServiceError(f"Submission failed "
+                                                        f"for {i}th chunk "
+                                                        f"of {t} "
+                                                        f"after {N_tries} attempts")
             try:
                 job = WISEDataBase.service.submit_job(qstring, uploads={'ids': Table(tab_d)})
                 job.run()
+
+                if isinstance(job.phase, type(None)):
+                    raise vo.dal.DALServiceError(f"Job submission failed. No phase!")
+
                 logger.info(f'submitted job for {t} for chunk {i}: ')
                 logger.debug(f'Job: {job.url}; {job.phase}')
                 self.tap_jobs[t][i] = job
                 self.queue.put((t, i))
                 break
+
             except (requests.exceptions.ConnectionError, vo.dal.exceptions.DALServiceError) as e:
                 wait = 60
                 N_tries -= 1
@@ -839,34 +870,34 @@ class WISEDataBase(abc.ABC):
              f"{self._split_chunk_key}{chunk_number}.csv"
         return os.path.join(self._cache_photometry_dir, fn)
 
+    @staticmethod
+    def _give_up_tap(e):
+        return ("Job is not active!" in str(e))
+
+    @backoff.on_exception(
+        backoff.expo,
+        vo.dal.exceptions.DALServiceError,
+        giveup=_give_up_tap,
+        max_tries=50,
+        on_backoff=backoff_hndlr
+    )
     def _thread_wait_and_get_results(self, t, i):
         logger.info(f"Waiting on {i}th query of {t} ........")
+
         _job = self.tap_jobs[t][i]
-        # Sometimes a connection Error occurs.
-        # In that case try again until job.wait() exits normally
-        _ntries = 10
-        while True:
-            try:
-                _job.wait()
-                break
-            except vo.dal.exceptions.DALServiceError as e:
-                msg = f"{i}th query of {t}: DALServiceError: {e}; trying again in 6 min"
-                if _ntries < 10:
-                    msg += f' ({_ntries} tries left)'
-
-                logger.warning(msg)
-                time.sleep(60 * 6)
-                if '404 Client Error: Not Found for url' in str(e):
-                    _ntries -= 1
-
+        _job.wait()
         logger.info(f'{i}th query of {t}: Done!')
+
         lightcurve = _job.fetch_result().to_table().to_pandas()
         fn = self._chunk_photometry_cache_filename(t, i)
         logger.debug(f"{i}th query of {t}: saving under {fn}")
+
         cols = dict(self.photometry_table_keymap[t]['mag'])
         cols.update(self.photometry_table_keymap[t]['flux'])
+
         if 'allwise' in t:
             cols['cntr_mf'] = 'allwise_cntr'
+
         lightcurve.rename(columns=cols).to_csv(fn)
         return
 
@@ -1428,15 +1459,14 @@ class WISEDataBase(abc.ABC):
         else:
             raise NotImplementedError
 
-    def _load_metadata(self, service, chunk_number=None, jobID=None, remove=False):
+    def _load_metadata(self, service, chunk_number=None, jobID=None, return_filename=False):
         fn = self._metadata_filename(service, chunk_number, jobID)
         try:
             logger.debug(f"loading {fn}")
             with open(fn, "r") as f:
                 metadata = json.load(f)
-            if remove:
-                logger.debug(f"removing")
-                os.remove(fn)
+            if return_filename:
+                return metadata, fn
             return metadata
         except FileNotFoundError:
             logger.warning(f"No file {fn}")
@@ -1507,19 +1537,26 @@ class WISEDataBase(abc.ABC):
         else:
             raise NotImplementedError
 
-        lcs = None
+        metadata = None
+        fns = list()
         for i in itr[1]:
             kw = dict(kwargs)
             kw[itr[0]] = i
-            kw['remove'] = remove
-            ilcs = self._load_metadata(**kw)
-            if not isinstance(ilcs, type(None)):
-                if isinstance(lcs, type(None)):
-                    lcs = dict(ilcs)
+            kw['return_filename'] = True
+            res = self._load_metadata(**kw)
+            if not isinstance(res, type(None)):
+                imetadata, ifn = res
+                fns.append(ifn)
+                if isinstance(metadata, type(None)):
+                    metadata = dict(imetadata)
                 else:
-                    lcs.update(ilcs)
+                    metadata.update(imetadata)
 
-        self._save_metadata(lcs, service=service, chunk_number=chunk_number, overwrite=overwrite)
+        self._save_metadata(metadata, service=service, chunk_number=chunk_number, overwrite=overwrite)
+
+        if remove:
+            for fn in tqdm.tqdm(fns, desc='removing files'):
+                os.remove(fn)
 
     @abc.abstractmethod
     def calculate_metadata_single(self, lcs):
